package com.technogumbo
{
	/*
	Copyright (c) 2011, Charles Palen
	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
    Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

	import flash.display.Sprite;
	import flash.events.Event;
	import flash.events.MouseEvent;
	import flash.events.NetStatusEvent;
	import flash.events.StatusEvent;
	import flash.events.TimerEvent;
	import flash.utils.Timer;
	
	import flash.media.Camera;
	import flash.media.Microphone;
	import flash.media.Video;
	import flash.net.NetConnection;
	import flash.net.NetStream;
	import flash.text.TextFieldAutoSize;
	
	// These imports come from the swc. They are not nativeley available to applications
	// built outside the Flash CS versions.
	import fl.controls.Button;
	import fl.controls.TextInput;
	import fl.controls.Label;
	import fl.controls.ComboBox;
	import fl.controls.CheckBox;
	import fl.data.DataProvider;
	
	/**
	 * ...
	 * @author Charles
	 * 
	 * This sample was created in July 2011 using
	 * SVN version of Red5 RC1/RC2. Don't criticize me on doing this all in one file.
	 * I did this on purpose. I've seen too many OO heavy examples that will really lose beginners.
	 * 
	 * The H.264 encoders in Flash Player 11 may have a great effect on the items I describe below
	 * I would assume they are going to utilize the same behavior.  Think about the tough task of
	 * encoding video in real time? H.264 implementations are pretty CPU intensive so it may
	 * effect the buffers even more.  In any case, this is how to do it before Flash Player 11.
	 * 
	 * Be aware that recording performance and behavior varries greatly
	 * depending on what version of Red5 you're using.
	 * 
	 * In RC1 - there is a configuration property in
	 * conf/red5-common.xml called queueThreshold
	 * By default this is 33, but according to AVChat and hdfvr, you should set this very high,
	 * like 3600
	 * http://avchathq.com/blog/recording-high-quality-flash-video-over-slow-internet-connections-part-3/
	 * http://hdfvr.com/documentation#11
	 * 
	 * 
	 * Also, if you include audio and video in the same NetStream, it will almost ALWAYS mess up the video
	 * or cause the video to drop frames because Flash player gives the audio transport priority and drops
	 * video frames.  In order to get around this, we record audio and video separateley, then play them
	 * back separeteley as well.  This causes slight audio out of sync issues some times, but its better than
	 * the trash quality video you normally get.
	 * 
	 * Also, be aware that at time of authoring, there is no way that I am aware of to record video in Flash/AIR/Flex without using
	 * a media server.
	 */
	public class Main extends Sprite 
	{
		private var activeCamera:Camera;
		private var activeMic:Microphone;
		private var activeConnection:NetConnection;
		private var activeStream:NetStream;
		private var audioOnlyStream:NetStream;
		
		private var onScreenVideo:Video;
		private var audioPlaybackVideo:Video;
		
		private var camAllowed:Boolean = false;
		private var micAllowed:Boolean = false;
		private var recordingHalted:Boolean = false;
		private var currentlyRecordedFileName:String = "";
		
		private var bufferCheckTimer:Timer;
		
		// UI Purposes - These Controls come from the swc
		private var btnRecord:Button;
		private var btnPlay:Button;
		private var btnStartMerge:Button;
		
		private var cbMicHz:ComboBox;
		private var lblMic:Label;
		
		private var lblVidPlaceholder:Label;
		private var lblVidResolution:Label;
		
		private var lblCamWidth:Label;
		private var txtCamWidth:TextInput;
		
		private var lblCamHeight:Label;
		private var txtCamHeight:TextInput;
		
		private var lblCamFPS:Label;
		private var txtCamFPS:TextInput;
		
		private var lblCamBandwidth:Label;
		private var txtCamBandwidth:TextInput;
		
		private var lblCamQuality:Label;
		private var txtCamQuality:TextInput;
		
		private var lblVideoBuffer:Label;
		private var lblAudioBuffer:Label;
		
		private var lblMediaSvr:Label;
		private var txtMediaSvrAdd:TextInput;
		private var btnMediaSrvr:Button;
		private var lblBufferSize:Label;
		
		private var lblNetStreamBuff:Label;
		private var txtNetStreamBuff:TextInput;
		
		private var lblNetStreamBehavior:Label;
		private var cbNetStreamBehavior:CheckBox;
		
		private var lblStatus:Label;
		
		public function Main():void 
		{
			if (stage) init();
			else addEventListener(Event.ADDED_TO_STAGE, init);
		}
		
		private function init(e:Event = null):void 
		{
			removeEventListener(Event.ADDED_TO_STAGE, init);
			// entry point
			
			// Setup the UI
			setupUI();
		}
		
		/**
		 * Sets up the UI.
		 * Don't get lost in here. The important parts of the example are definatley not this.
		**/
		private function setupUI():void {
			
			onScreenVideo = new Video(320, 240);
			onScreenVideo.smoothing = true;
			onScreenVideo.x = 10;
			onScreenVideo.y = 10;
			
			// We smash the audio only video behind the real video that contains..well video. Audio video is only used for
			// playing back audio.
			
			// Make sure to set the video to the same size as your camera!
			// NAH - making it smaller and it will just get crushed
			audioPlaybackVideo = new Video(320, 240);
			audioPlaybackVideo.x = 10;
			audioPlaybackVideo.y = 10;
			
			addChild(audioPlaybackVideo); // -- This is directly behind the onScreenVideo so its not visible
			addChild(onScreenVideo);
			
			lblVidPlaceholder = new Label();
			lblVidPlaceholder.text = "Once You Record Video Wideo Will Be Displayed Here";
			lblVidPlaceholder.width = 200;
			lblVidPlaceholder.height = 100;
			lblVidPlaceholder.wordWrap = true;
			lblVidPlaceholder.x = 10;
			lblVidPlaceholder.y = 100;
			addChild(lblVidPlaceholder);
			
			lblMediaSvr = new Label();
			lblMediaSvr.text = "Media Server Application URI:";
			lblMediaSvr.width = 200;
			lblMediaSvr.x = 340;
			lblMediaSvr.y = 10;
			addChild(lblMediaSvr);
			
			txtMediaSvrAdd = new TextInput();
			txtMediaSvrAdd.text = "rtmp://localhost/vod";
			txtMediaSvrAdd.x = 340;
			txtMediaSvrAdd.y = 30;
			txtMediaSvrAdd.width = 160;
			addChild(txtMediaSvrAdd);
			
			btnMediaSrvr = new Button();
			btnMediaSrvr.label = "Connect";
			btnMediaSrvr.addEventListener(MouseEvent.CLICK, handleMedSrvrBtn, false, 0, true);
			btnMediaSrvr.x = 520;
			btnMediaSrvr.y = 30;
			addChild(btnMediaSrvr);
			
			lblVidResolution = new Label();
			lblVidResolution.text = "Settings Take Effect When You Connect to the Media Server In This Example. Disconnect then Re-connect to Change Camera and Microphone Settings.";
			lblVidResolution.wordWrap = true;
			lblVidResolution.width = 300;
			lblVidResolution.height = 200;
			//lblVidResolution.autoSize = TextFieldAutoSize.LEFT;
			lblVidResolution.x = 340;
			lblVidResolution.y = 60;
			addChild(lblVidResolution);
			
		
			lblCamWidth = new Label();
			lblCamWidth.text = "Width:";
			lblCamWidth.x = 340;
			lblCamWidth.y = 120;
			addChild(lblCamWidth);
			
			txtCamWidth = new TextInput();
			txtCamWidth.width = 35;
			txtCamWidth.height = 20;
			txtCamWidth.x = 375;
			txtCamWidth.y = 120;
			txtCamWidth.text = "640";
			addChild(txtCamWidth);
			
			lblCamHeight = new Label();
			lblCamHeight.text = "Height:";
			lblCamHeight.y = 120;
			lblCamHeight.x = 435;
			addChild(lblCamHeight);
			
			txtCamHeight = new TextInput();
			txtCamHeight.width = 35;
			txtCamHeight.height = 20;
			txtCamHeight.y = 120;
			txtCamHeight.x = 470;
			txtCamHeight.text = "480";
			addChild(txtCamHeight);
			
			lblCamFPS = new Label();
			lblCamFPS.text = "FPS:";
			lblCamFPS.y = 120;
			lblCamFPS.x = 525;
			addChild(lblCamFPS);
			
			txtCamFPS = new TextInput();
			txtCamFPS.width = 35;
			txtCamFPS.height = 20;
			txtCamFPS.x = 550;
			txtCamFPS.y = 120;
			txtCamFPS.text = "29.97";
			addChild(txtCamFPS);
			
			lblCamBandwidth = new Label();
			lblCamBandwidth.text = "Bandwidth:";
			lblCamBandwidth.x = 340;
			lblCamBandwidth.y = 150;
			lblCamBandwidth.width = 60;
			addChild( lblCamBandwidth );
			
			txtCamBandwidth = new TextInput();
			txtCamBandwidth.width = 30;
			txtCamBandwidth.height = 20;
			txtCamBandwidth.y = 150;
			txtCamBandwidth.x = 400;
			txtCamBandwidth.text = "0";
			addChild( txtCamBandwidth );
			
			lblCamQuality = new Label();
			lblCamQuality.text = "Quality:";
			lblCamQuality.x = 435;
			lblCamQuality.y = 150;
			lblCamQuality.width = 65;
			addChild( lblCamQuality );
			
			txtCamQuality = new TextInput();
			txtCamQuality.width = 30;
			txtCamQuality.height = 20;
			txtCamQuality.y = 150;
			txtCamQuality.x = 470;
			txtCamQuality.text = "85";
			addChild( txtCamQuality );
			
			lblMic = new Label();
			lblMic.text = "Microphone Record Rate";
			lblMic.width = 150;
			lblMic.x = 340;
			lblMic.y = 180;
			addChild(lblMic);
			
			var sampleRates:Array = [ 
			{label:"44kHz", data:"44"}, 
			{label:"22kHz", data:"22"}, 
			{label:"11kHz", data:"11" },
			{label:"8kHz", data:"8"}, 
			{label:"5kHz", data:"5"}
			]; 
			
			cbMicHz = new ComboBox();
			cbMicHz.dropdownWidth = 100;
			cbMicHz.move(340, 200);
			cbMicHz.dataProvider = new DataProvider(sampleRates);
			addChild(cbMicHz);
		
			lblNetStreamBuff = new Label();
			lblNetStreamBuff.text = "Netstream Buffer Time During Recording (Automatically Set to 1 For Playback)";
			lblNetStreamBuff.x = 340;
			lblNetStreamBuff.y = 240;
			lblNetStreamBuff.width = 200;
			lblNetStreamBuff.height = 60;
			lblNetStreamBuff.wordWrap = true;
			addChild( lblNetStreamBuff );
			
			txtNetStreamBuff = new TextInput();
			txtNetStreamBuff.text = "60";
			txtNetStreamBuff.height = 20;
			txtNetStreamBuff.width = 30;
			txtNetStreamBuff.x = 340;
			txtNetStreamBuff.y = 280;
			addChild( txtNetStreamBuff );
			
			lblNetStreamBehavior = new Label();
			lblNetStreamBehavior.text = "Netstream Waits For Buffer To Empty Before Calling close?";
			lblNetStreamBehavior.width = 200;
			lblNetStreamBehavior.height = 60;
			lblNetStreamBehavior.wordWrap = true;
			lblNetStreamBehavior.x = 340;
			lblNetStreamBehavior.y = 320;
			addChild( lblNetStreamBehavior );
			
			cbNetStreamBehavior = new CheckBox();
			cbNetStreamBehavior.x = 340;
			cbNetStreamBehavior.y = 360;
			cbNetStreamBehavior.selected = true;
			cbNetStreamBehavior.label = "";
			addChild( cbNetStreamBehavior );
			
			// -- Left side UI
			btnRecord = new Button();
			btnRecord.label = "Record";
			btnRecord.addEventListener(MouseEvent.CLICK, handleRecordStartStop, false, 0, true);
			btnRecord.x = 10;
			btnRecord.y = 290;
			addChild(btnRecord);
			btnRecord.enabled = false;
			
			btnPlay = new Button();
			btnPlay.label = "Play Recording";
			btnPlay.addEventListener(MouseEvent.CLICK, handlePlaybackClick, false, 0, true);
			btnPlay.x = 200;
			btnPlay.y = 290;
			addChild(btnPlay);
			btnPlay.enabled = false;
			
			btnStartMerge = new Button();
			btnStartMerge.label = "Start Merge";
			btnStartMerge.addEventListener(MouseEvent.CLICK, handleMediaServerMergeCall, false, 0, true);
			btnStartMerge.x = 10;
			btnStartMerge.y = 320;
			addChild( btnStartMerge );
			btnStartMerge.enabled = false;
			
			lblVideoBuffer = new Label();
			lblVideoBuffer.text = "Video Stream Buffer: 0";
			lblVideoBuffer.x = 10;
			lblVideoBuffer.y = 250;
			lblVideoBuffer.width = 320;
			addChild(lblVideoBuffer);
			
			lblAudioBuffer = new Label();
			lblAudioBuffer.text = "Audio Stream Buffer: 0";
			lblAudioBuffer.x = 10;
			lblAudioBuffer.y = 270;
			lblAudioBuffer.width = 320;
			addChild(lblAudioBuffer);
			
			lblStatus = new Label();
			lblStatus.x = 10;
			lblStatus.y = 360;
			lblStatus.width = 320;
			lblStatus.height = 120;
			lblStatus.wordWrap = true;
			setStatus("Setting up UI:");
			addChild(lblStatus);
			
			
		}
		
		/**
		 * Just a utility function for the UI
		 * @param	_OnOff
		 */
		private function disableEnableRecordingSettings(_OnOff:Boolean):void {
			if (_OnOff == true) {
				
				txtMediaSvrAdd.enabled = true;
				txtCamWidth.enabled = true;
				txtCamHeight.enabled = true;
				txtCamFPS.enabled = true;
				txtCamBandwidth.enabled = true;
				txtCamQuality.enabled = true;
				cbMicHz.enabled = true;
				txtNetStreamBuff.enabled = true;
				cbNetStreamBehavior.enabled = true;
				
			} else {
				
				txtMediaSvrAdd.enabled = false;
				txtCamWidth.enabled = false;
				txtCamHeight.enabled = false;
				txtCamFPS.enabled = false;
				txtCamBandwidth.enabled = false;
				txtCamQuality.enabled = false;
				cbMicHz.enabled = false;
				txtNetStreamBuff.enabled = false;
				cbNetStreamBehavior.enabled = false;
				
			}
		}
		
		/**
		 * Another utility function for the UI. Changes the status text in the lower left
		 * @param	_NewStatus
		 */
		private function setStatus(_NewStatus:String):void {
			lblStatus.text = "STATUS: " + _NewStatus;
		}
		
		/**
		 * Mouse Click Event button handler for connect/Disconnect to the media server
		 * (The button in the top left)
		 * @param	e
		 */
		private function handleMedSrvrBtn(e:MouseEvent):void {
			if (btnMediaSrvr.label == "Connect") {
			
				// Disable UI recording settings
				disableEnableRecordingSettings(false);
				
				setupConnection();
				
				btnMediaSrvr.label = "Disconnect";
			} else {
				
				// Destroy the camera
				destroyBufferCheckTimer();
				// Destroy camera and mic
				destroyCameraAndMic();
				// Destroy any leftover or active streams
				destroyStreams();
				// Destroy the original connection
				destroyConnection();
				
				// Enable UI recording settings
				disableEnableRecordingSettings(true);
				
				btnRecord.enabled = false;
				btnPlay.enabled = false;
				btnStartMerge.enabled = false;
				
				btnMediaSrvr.label = "Connect";
			}
		}

		/**
		 * The mouse click event handler for the recording start/stop button
		 * @param	e
		 */
		private function handleRecordStartStop(e:MouseEvent):void {
			if (btnRecord.label == "Record") {
				
				recordingHalted = false;
				
				lblVidPlaceholder.visible = false;
				setupStream();
				setupCameraAndMic();
					
				btnPlay.label = "Play Recording";
				btnPlay.enabled = false;
				btnStartMerge.enabled = false;
				
				btnRecord.label = "Stop";
			} else {
				
				if (activeStream != null) {
					setStatus("Done recording, waiting for buffers to empty.");
					
					activeStream.attachCamera(null);
					activeStream.attachAudio(null);
					
					audioOnlyStream.attachCamera(null);
					audioOnlyStream.attachAudio(null);
					
					recordingHalted = true;
					
					// Looks like we aren't waiting for the buffer to empty. Good luck!
					if (cbNetStreamBehavior.selected == false) {
						destroyBufferCheckTimer();
						activeStream.close();
						audioOnlyStream.close();
						
						setStatus("We didn't wait for the buffers to empty and immediateley closed.");
						// OK - Enable Playback - playback time
						btnPlay.enabled = true;
						// Enable calling 
						btnStartMerge.enabled = true;
					}
				}
				
				btnRecord.label = "Record";
			}
		}
		
		/**
		 * The mouse click handler for the button that allows you to play back a recording.
		 * @param	e
		 */
		private function handlePlaybackClick(e:MouseEvent):void {
			if (btnPlay.label == "Play Recording") {
				
				doRecordingPlayback();
				
				btnPlay.label = "Stop";
			} else {
				
				stopRecordingPlayback();
				
				btnPlay.label = "Play Recording";
			}
		}
		
		/**
		 * Mouse event click handler. This will attempt to do a netconnection.call to 
		 * call a function on the media server. This was intended for use with the
		 * recording_merger server-side application included in the same tutorial as this application.
		 * 
		 * Im ignoring a responder here because the server-side function is void and returns nothing.
		 * In the recording_merger application it actually invokes a call back on this application that
		 * has nothing to do with a responder.
		 * 
		 * In the red5 server side application, I'm attempting to call a public function here.
		 * @param	e
		 */
		private function handleMediaServerMergeCall(e:MouseEvent):void {
			if (activeConnection != null) {
				activeConnection.call( "initiateTranscoder", null, currentlyRecordedFileName );
			}
		}
		
		/**
		 * This function is envoked by the media server when transcoding completes..if using the 
		 * recording_merger application.
		 * 
		 * @param	_FullFilePath
		 */
		public function transcodingCallback(_FullFilePath:String):void {
			setStatus("transcodingCallback Final Video at: " + _FullFilePath);
		}
		
		/**
		 * Sets up the Netconnection to the media server
		 */
		private function setupConnection():void {
			if(activeConnection == null) {
				activeConnection = new NetConnection();
				// Setting the client to this so when the server attempts to call functions on the client
				// it will go to this class.
				activeConnection.client = this;
				activeConnection.addEventListener(NetStatusEvent.NET_STATUS, handleConnectionStatus, false, 0, true);
				activeConnection.connect( txtMediaSvrAdd.text );
				setStatus("Connecting to: " + txtMediaSvrAdd.text );
			} else {
				destroyConnection(true);
			}
		}
		
		/**
		 * A utility function for destroying the active netconnection object.
		 * 
		 * Please ignore the optionalCallback items, they are just a methodology I use
		 * @param	_OptionalCallback
		 */
		private function destroyConnection(_OptionalCallback:Boolean = false):void {
			if (activeConnection != null) {
				
				activeConnection.close();
				activeConnection.removeEventListener(NetStatusEvent.NET_STATUS, handleConnectionStatus);
				activeConnection = null;
				
				if (_OptionalCallback == true) {
					setupConnection();
				}
			}
		}
		
		/**
		 * The NetConnections netstatus event handler.  This will fire when a connection is closed
		 * or if you successfully connect to a media server.
		 * @param	e
		 */
		private function handleConnectionStatus(e:NetStatusEvent):void {
			trace("handleConnectionStatus - " + e.info.code );
			switch(e.info.code) {
				case 'NetConnection.Connect.Success':
					setStatus("Successfully connected to: " + txtMediaSvrAdd.text );
					btnRecord.enabled = true;
					
				break;
				case 'NetConnection.Connect.Closed':
					setStatus("Netconnection Closed to: " + txtMediaSvrAdd.text );
				break;
			}
		}
		
		/**
		 * This function begins setup on the audio and video netstream which are part of the
		 * core of recording and playback when dealing with a media server.
		 */
		private function setupStream():void {
			if (activeStream == null) {
				activeStream = new NetStream(activeConnection);
				audioOnlyStream = new NetStream(activeConnection);
				
				// Set the client object so the netstreams can call onMetaData when we do playback
				activeStream.client = this;
				audioOnlyStream.client = this;
				
				activeStream.addEventListener(NetStatusEvent.NET_STATUS, handleStreamStatus, false, 0, true);
				audioOnlyStream.addEventListener(NetStatusEvent.NET_STATUS, handleAudioOnlyStreamStatus, false, 0, true);
				// Set the buffer time real high so the player doesnt just start dropping frames all over the place
				// to preserve the tiny default .1 buffer
				// http://livedocs.adobe.com/flash/9.0/ActionScriptLangRefV3/flash/net/NetStream.html#bufferTime
				activeStream.bufferTime = Number( txtNetStreamBuff.text );
				audioOnlyStream.bufferTime = Number( txtNetStreamBuff.text );
			} else {
				destroyStreams(true);
			}
		}
		
		/**
		 * Just a function for cleaning up any of the Netstreams as completley as possible.
		 * Again, ignore the optional callback stuff. Its just a methodology I use in OO programming.
		 * @param	_OptionalCallback
		 */
		private function destroyStreams(_OptionalCallback:Boolean = false):void {
			if (activeStream != null) {
				
				if(onScreenVideo != null) {
					onScreenVideo.attachNetStream(null);
				}
				if(audioPlaybackVideo != null) {
					audioPlaybackVideo.attachNetStream(null);
				}
			
				activeStream.attachCamera(null);
				activeStream.attachAudio(null);
				
				audioOnlyStream.attachCamera(null);
				audioOnlyStream.attachAudio(null);
					
				activeStream.close();
				audioOnlyStream.close();
				
				activeStream.removeEventListener(NetStatusEvent.NET_STATUS, handleStreamStatus);
				audioOnlyStream.removeEventListener(NetStatusEvent.NET_STATUS, handleAudioOnlyStreamStatus);
				
				activeStream = null;
				audioOnlyStream = null;
				
				if (_OptionalCallback == true) {
					setupStream();
				}
			}
		}
	
		// When streaming video or doing playback using netstream you always have to setup onMetaData
		// so this is used for both AUDIO and VIDEO
		// Where we set the netstream's client to this, it allows the netstream to automatically call this function
		public function onMetaData(info:Object):void {
			trace("playback called onMetaData");
			setStatus("Playback of recording called onMetaData");
		}
		
		// Also a callback needed when streaming video back
		// http://livedocs.adobe.com/flash/9.0/ActionScriptLangRefV3/flash/net/NetStream.html#event:onPlayStatus
		public function onPlayStatus(info:Object):void {
			trace("playback called onPlayStatus");
			setStatus("Playback of recording called onPlayStatus");
		}
		
		/**
		 * Handles NetStatus events on the video NetStream
		 * We dont particularly do anything here, but these are real helpful
		 * if making a full video player.
		 * @param	e
		 */
		private function handleStreamStatus(e:NetStatusEvent):void {
			switch(e.info.code) {
				case 'NetStream.Buffer.Empty':
					setStatus("Video Netstream Buffer Empty");
				break;
				case 'NetStream.Buffer.Full':
					setStatus("Video Netstream Buffer Full");
				break;
				case 'NetStream.Buffer.Flush':
					setStatus("Video Netstream Buffer Flushed!!!!");
				break;
			}
		}
		
		/**
		 * Handles NetStatus events on the audio NetStream
		 * We dont particularly do anything here, but these are real helpful
		 * if making a full video player.
		 * @param	e
		 */
		private function handleAudioOnlyStreamStatus(e:NetStatusEvent):void {
			switch(e.info.code) {
				case 'NetStream.Buffer.Empty':
					setStatus("Audio Netstream Buffer Empty");
				break;
				case 'NetStream.Buffer.Full':
					setStatus("Audio Netstream Buffer Full");
				break;
				case 'NetStream.Buffer.Flush':
					setStatus("Audio Netstream Buffer Flushed!!!");
				break;
			}
		}
		
		/**
		 * Prepares the camera and microphone based on the parameters in the UI.
		 */
		private function setupCameraAndMic():void {
			if (activeCamera == null) {
				// Gets the default camera on the system - you can pass "0", "1", etc to manually select devices
				activeCamera = Camera.getCamera();
				activeCamera.addEventListener(StatusEvent.STATUS, handleCameraStatus, false, 0, true);
				// 320x240 at 20fps
				activeCamera.setMode( int( txtCamWidth.text ) , int( txtCamHeight.text ) , Number( txtCamFPS.text ) );
				// Youll probably have to set quality lower depending on your webcam.  All webcams transmit different
				// ammounts of data. Theyll get the buffer so high it's impractical for users to wait that long!
				activeCamera.setQuality( int( txtCamBandwidth.text ) , int( txtCamQuality.text ) );
				
				activeMic = Microphone.getMicrophone();
				activeMic.addEventListener(StatusEvent.STATUS, handleMicrophoneStatus, false, 0, true);
				// Rate in khz
				// http://livedocs.adobe.com/flash/9.0/ActionScriptLangRefV3/flash/media/Microphone.html#rate
				activeMic.rate = int( cbMicHz.selectedItem.data  );
				activeMic.gain = 50;
				// Audio wont transmit all the time causing breaks in recorded audio stream screwing up the
				// flv.  Make sure even if its silence while recording the mic is always sending data
				activeMic.setSilenceLevel(0);
				
				
				onScreenVideo.attachCamera(activeCamera);
				if (activeStream != null) {
					activeStream.attachCamera(activeCamera);
				}
				if (audioOnlyStream != null) {
					audioOnlyStream.attachAudio(activeMic);
				}
				
				// Wait for the security settings - If we have already approved these - the fun little dialog wont ever show up again,
				// and the events on camera and mic wont fire, so just check initially.
				setStatus("Waiting for camera and microphone security settings.");
				
				// -- Will just proceede to publish recording if we already have passed security settings
				finishCamAndMicSetup();
			} else {
				destroyCameraAndMic(true);
			}
		}
		
		/**
		 * A utility function for completley cleaning up the camrea and microphone
		 * @param	_OptionalCallback
		 */
		private function destroyCameraAndMic(_OptionalCallback:Boolean = false):void {
			if (activeCamera != null) {
				
				if ( onScreenVideo != null) {
					onScreenVideo.attachCamera(null);
				}
				
				if (activeStream != null) {
					activeStream.attachCamera(null);
				}
				
				if (audioOnlyStream != null) {
					audioOnlyStream.attachAudio(null);
				}
				
				activeCamera.removeEventListener(StatusEvent.STATUS, handleCameraStatus);
				activeCamera = null;
				
				activeMic.removeEventListener(StatusEvent.STATUS, handleMicrophoneStatus);
				activeMic = null;
				
				// Just in case - will do nothing if already destroyed
				destroyBufferCheckTimer();
				
				if (_OptionalCallback == true) {
					setupCameraAndMic();
				}
			}
		}
		
		/**
		 * We have to wait to actually start recording due to the Flash Player security
		 * dialog. So this method is called only after we have the OK on security settings.
		 */
		private function finishCamAndMicSetup():void {
			if (camAllowed == true && micAllowed == true) {
				if (activeStream != null) {
					
					setStatus("Recording....");
									
					// This will start recording - the name we specify here is the name of the final file
					// on the server.
					// I'll use the data functionality to generate unique names
					var tempDate:Date = new Date();
					var uniqueFileName:String = "RecordTest_" + String(tempDate.getMinutes()) + String(tempDate.getMilliseconds());
					
					currentlyRecordedFileName = uniqueFileName;
					
					// OK lets start recording - look in the "streams" directory inside of red5
					// inside of the app we connected to in the NetConnection
					activeStream.publish(uniqueFileName + "_Video", "record");
					audioOnlyStream.publish(uniqueFileName + "_Audio", "record");
					
					createBufferCheckTimer();
				}
			}
		}
		
		/**
		 * Utility function for creating the timer that checks the netstreams buffer.
		 * This timer executes every 100 milliseconds, or 10 times a second because 1000ms = 1 second
		 */
		private function createBufferCheckTimer():void {
			if (bufferCheckTimer == null) {
					bufferCheckTimer = new Timer(100);
					bufferCheckTimer.addEventListener(TimerEvent.TIMER, handleBufferCheck, false, 0, true);
					bufferCheckTimer.start();
			} else {
				destroyBufferCheckTimer(true);
			}
		}
		
		/**
		 * Utility function for destroying the buffer check timer.
		 * @param	_OptionalCallback
		 */
		private function destroyBufferCheckTimer(_OptionalCallback:Boolean = false):void {
			if (bufferCheckTimer != null) {
					bufferCheckTimer.stop();
					bufferCheckTimer.removeEventListener(TimerEvent.TIMER, handleBufferCheck);
					bufferCheckTimer = null;
					
					if (_OptionalCallback == true) {
						createBufferCheckTimer();
					}
			}
		}
		
		/**
		 * Whenever the timer runs, this function is called, so its called 10 times a second
		 * to check the state of the buffers when recording.
		 * @param	e
		 */
		private function handleBufferCheck(e:TimerEvent):void {
			if (activeStream != null) {
				
			    if (lblAudioBuffer != null) {
					lblAudioBuffer.text = "Audio Stream Buffer: " + String( audioOnlyStream.bufferLength );
				}

				if (lblVideoBuffer != null) {
					lblVideoBuffer.text = "Video Stream Buffer: " + String( activeStream.bufferLength );
				}
				
				if (recordingHalted == true) {
					if ( (activeStream.bufferLength == 0) && (audioOnlyStream.bufferLength == 0) ) {
						activeStream.close();
						audioOnlyStream.close();
						
						bufferCheckTimer.stop();
						bufferCheckTimer.removeEventListener(TimerEvent.TIMER, handleBufferCheck);
						bufferCheckTimer = null;
						
						setStatus("Buffers Empty. Press 'Play Recording' to start playback.");
						// OK - Enable Playback - playback time
						btnPlay.enabled = true;
						btnStartMerge.enabled = true;
					}
				}
			}
			
			if (bufferCheckTimer != null) {
				bufferCheckTimer.reset();
				bufferCheckTimer.start();
			}
		}
		
		/**
		 * The flash player security dialog will prompt users for camera access when we request the camera
		 * use this event to figure out if they allowed it or not.
		 * http://livedocs.adobe.com/flash/9.0/ActionScriptLangRefV3/flash/media/Camera.html#getCamera%28%29
		 * @param	e
		 */
		private function handleCameraStatus(e:StatusEvent):void {
			trace("handleCameraStatus - " + e.code);
			switch(e.code) {
				case 'Camera.muted':
					// Show a message
					setStatus("Camera muted - unable to continue");
				break;
				case 'Camera.Unmuted':
					camAllowed = true;
					finishCamAndMicSetup();
				break;
			}
		}
		
		/**
		 * The flash player security dialog will prompt users for microphone access when we request the camera
		 * use this event to figure out if they allowed it or not.
		 * http://livedocs.adobe.com/flash/9.0/ActionScriptLangRefV3/flash/media/Camera.html#getCamera%28%29
		 * @param	e
		 */
		private function handleMicrophoneStatus(e:StatusEvent):void {
			trace("handleMicrophoneStatus - " + e.code);
			switch(e.code) {
				case 'Microphone.Muted':
					// Show a message
					setStatus("Microphone muted - unable to continue");
				break;
				case 'Microphone.Unmuted':
					micAllowed = true;
					finishCamAndMicSetup();
				break;
			}
		}
		
		/**
		 * When playing back video/audio, we can use the same Netstreams we used to record
		 * as well as the same video.
		 * 
		 * Take special note of the bufferTime change here.  If I left it at 60 or whatever was set
		 * in the UI, the netstreams try and download that much time of video/audio before starting
		 * playback.
		 * 
		 * I think the default value on netstreams of .1 sec is really intended for playback because
		 * so many people do playback vs video recording.
		 */
		private function doRecordingPlayback():void {
			setStatus("Starting simultanious playback of recording audio and video");
			
			onScreenVideo.attachCamera(null);
			onScreenVideo.attachNetStream(activeStream);
			audioPlaybackVideo.attachNetStream(audioOnlyStream);
			
			// Attempt to play back the audio and video file simultaniously
			// First, adjust the buffers as this will be BAD for playback
			activeStream.bufferTime = 1;
			audioOnlyStream.bufferTime = 1;
			
			activeStream.play(currentlyRecordedFileName + "_Video");
			audioOnlyStream.play(currentlyRecordedFileName + "_Audio");
		}
		
		/**
		 * If for some reason someone stops the playback. I just pause the audio and
		 * video stream.  they get destroyed when we record again anyway.
		 */
		private function stopRecordingPlayback():void {
			setStatus("Stopping Recording Playback");
			
			if ( activeStream != null) {
				activeStream.pause();
			}
			if ( audioOnlyStream != null) {
				audioOnlyStream.pause();
			}
		}
		
	}
	
}