package com.technogumbo
{
	import adobe.utils.CustomActions;
	import flash.display.Sprite;
	import flash.events.Event;
	import flash.events.MouseEvent;
	import flash.events.NetStatusEvent;
	import flash.events.StatusEvent;
	import flash.events.TimerEvent;
	import flash.utils.Timer;
	
	import flash.media.Camera;
	import flash.media.Microphone;
	import flash.media.Video;
	import flash.net.NetConnection;
	import flash.net.NetStream;

	// FP 11+ only
	import flash.media.H264VideoStreamSettings;
	import flash.media.H264Profile;
	import flash.media.H264Level;
	import flash.media.VideoStreamSettings;
	import flash.media.VideoCodec;

	/**
	 * ...
	 * @author Charles
	 * 
	 * THIS ONE USES Flex 4.5 + FP 11.0 Incubator to do H.264 encoding!!!
	 * 
	 * This sample was created in April 2011 using
	 * SVN version of Red5 RC1
	 * 
	 * Be aware that recording performance and behavior varries greatly
	 * depending on what version of Red5 you're using.
	 * 
	 * In RC1 - there is a configuration property in
	 * conf/red5-common.xml called queueThreshold
	 * By default this is 33, but according to AVChat and hdfvr, you should set this very high,
	 * like 3600
	 * http://avchathq.com/blog/recording-high-quality-flash-video-over-slow-internet-connections-part-3/
	 * http://hdfvr.com/documentation#11
	 * 
	 * 
	 * Also, if you include audio and video in the same NetStream, it will almost ALWAYS mess up the video
	 * or cause the video to drop frames because Flash player gives the audio transport priority and drops
	 * video frames.  In order to get around this, we record audio and video separateley, then play them
	 * back separeteley as well.  This causes slight audio out of sync issues some times, but its better than
	 * the trash quality video you normally get.
	 */
	public class Main extends Sprite 
	{
		private var activeCamera:Camera;
		private var activeMic:Microphone;
		private var activeConnection:NetConnection;
		private var activeStream:NetStream;
		private var audioOnlyStream:NetStream;
		
		private var onScreenVideo:Video;
		private var audioPlaybackVideo:Video;
		
		private var camAllowed:Boolean = false;
		private var micAllowed:Boolean = false;
		private var recordingHalted:Boolean = false;
		private var currentlyRecordedFileName:String = "";
		
		private var bufferCheckTimer:Timer;
		
		public function Main():void 
		{
			if (stage) init();
			else addEventListener(Event.ADDED_TO_STAGE, init);
		}
		
		private function init(e:Event = null):void 
		{
			removeEventListener(Event.ADDED_TO_STAGE, init);
			// entry point
			setupConnection();
			// Listen
			stage.addEventListener(MouseEvent.CLICK, handleAllClick, false, 0, true);
		}
		
		private function setupConnection():void {
			if(activeConnection == null) {
				activeConnection = new NetConnection();
				activeConnection.addEventListener(NetStatusEvent.NET_STATUS, handleConnectionStatus, false, 0, true);
				activeConnection.connect("rtmp://localhost/vod");
			}
		}
		
		private function handleConnectionStatus(e:NetStatusEvent):void {
			trace("handleConnectionStatus - " + e.info.code );
			switch(e.info.code) {
				case 'NetConnection.Connect.Success':
					setupStream();
					setupCameraAndMic();
				break;
				case 'NetConnection.Connect.Closed':
				
				break;
			}
		}
		
		private function setupStream():void {
			if (activeStream == null) {
				activeStream = new NetStream(activeConnection);
				audioOnlyStream = new NetStream(activeConnection);
				
				// Set the client object so the netstreams can call onMetaData when we do playback
				activeStream.client = this;
				audioOnlyStream.client = this;
				
				activeStream.addEventListener(NetStatusEvent.NET_STATUS, handleStreamStatus, false, 0, true);
				audioOnlyStream.addEventListener(NetStatusEvent.NET_STATUS, handleAudioOnlyStreamStatus, false, 0, true);
				// Set the buffer time real high so the player doesnt just start dropping frames all over the place
				// to preserve the tiny default .1 buffer
				// http://livedocs.adobe.com/flash/9.0/ActionScriptLangRefV3/flash/net/NetStream.html#bufferTime
				activeStream.bufferTime = 60;
				audioOnlyStream.bufferTime = 60;
				
				
				// FP 11 Only
				// H.264
				// http://sonnati.wordpress.com/2011/05/14/a-dream-comes-true-h-264-encoding-into-flash-player-11/
				 var h264Settings:H264VideoStreamSettings = new H264VideoStreamSettings();
				h264Settings.setProfileLevel(H264Profile.BASELINE, H264Level.LEVEL_2);
				activeStream.videoStreamSettings = h264Settings;
			}
		}
		
		// When streaming video or doing playback using netstream you always have to setup onMetaData
		// so this is used for both AUDIO and VIDEO
		// Where we set the netstream's client to this, it allows the netstream to automatically call this function
		public function onMetaData(info:Object):void {
			trace("playback called onMetaData");
		}
		
		// Also a callback needed when streaming video back
		// http://livedocs.adobe.com/flash/9.0/ActionScriptLangRefV3/flash/net/NetStream.html#event:onPlayStatus
		public function onPlayStatus(info:Object):void {
			trace("playback called onPlayStatus");
		}
		
		private function handleStreamStatus(e:NetStatusEvent):void {
			switch(e.info.code) {
				case 'NetStream.Buffer.Empty':
				
				break;
				case 'NetStream.Buffer.Full':
				
				break;
				case 'NetStream.Buffer.Flush':
				
				break;
			}
		}
		
		private function handleAudioOnlyStreamStatus(e:NetStatusEvent):void {
			switch(e.info.code) {
				case 'NetStream.Buffer.Empty':
				
				break;
				case 'NetStream.Buffer.Full':
				
				break;
				case 'NetStream.Buffer.Flush':
				
				break;
			}
		}
		
		private function setupCameraAndMic():void {
			if (activeCamera == null) {
				// Gets the default camera on the system - you can pass "0", "1", etc to manually select devices
				activeCamera = Camera.getCamera();
				activeCamera.addEventListener(StatusEvent.STATUS, handleCameraStatus, false, 0, true);
				// 320x240 at 20fps
				activeCamera.setMode(640, 480, 20);
				// Youll probably have to set quality lower depending on your webcam.  All webcams transmit different
				// ammounts of data. Theyll get the buffer so high it's impractical for users to wait that long!
				activeCamera.setQuality(0, 95);
				
				activeMic = Microphone.getMicrophone();
				activeMic.addEventListener(StatusEvent.STATUS, handleMicrophoneStatus, false, 0, true);
				// Rate in khz
				// http://livedocs.adobe.com/flash/9.0/ActionScriptLangRefV3/flash/media/Microphone.html#rate
				activeMic.rate = 44;
				activeMic.gain = 50;
				// Audio wont transmit all the time causing breaks in recorded audio stream screwing up the
				// flv.  Make sure even if its silence while recording the mic is always sending data
				activeMic.setSilenceLevel(0);
						
				// We smash the audio only video behind the real video that contains..well video. Audio video is only used for
				// playing back audio.
				
				// Make sure to set the video to the same size as your camera!
				// NAH - making it smaller and it will just get crushed
				audioPlaybackVideo = new Video(320, 240);
				addChild(audioPlaybackVideo); // -- This is directly behind the onScreenVideo so its not visible
				
				onScreenVideo = new Video(320, 240);
				onScreenVideo.smoothing = true;
				addChild(onScreenVideo);
				
				
				onScreenVideo.attachCamera(activeCamera);
				if (activeStream != null) {
					activeStream.attachCamera(activeCamera);
				}
				if (audioOnlyStream != null) {
					audioOnlyStream.attachAudio(activeMic);
				}
				// Wait for the security settings

			}
		}
		
		private function finishCamAndMicSetup():void {
			if (camAllowed == true && micAllowed == true) {
				
				if (activeStream != null) {
					
					// This will start recording - the name we specify here is the name of the final file
					// on the server.
					// I'll use the data functionality to generate unique names
					var tempDate:Date = new Date();
					var uniqueFileName:String = "RecordTest_" + String(tempDate.getMinutes()) + String(tempDate.getMilliseconds());
					
					currentlyRecordedFileName = uniqueFileName;
					
					// OK lets start recording - look in the "streams" directory inside of red5
					// inside of the app we connected to in the NetConnection
					activeStream.publish(uniqueFileName + "_Video", "record");
					audioOnlyStream.publish(uniqueFileName + "_Audio", "record");
					
					bufferCheckTimer = new Timer(100);
					bufferCheckTimer.addEventListener(TimerEvent.TIMER, handleBufferCheck, false, 0, true);
					bufferCheckTimer.start();
				}
			}
		}
		
		// The flash player security dialog will prompt users for camera access when we request the camera
		// use this event to figure out if they allowed it or not.
		// http://livedocs.adobe.com/flash/9.0/ActionScriptLangRefV3/flash/media/Camera.html#getCamera%28%29
		private function handleCameraStatus(e:StatusEvent):void {
			trace("handleCameraStatus - " + e.code);
			switch(e.code) {
				case 'Camera.muted':
					// Show a message
				break;
				case 'Camera.Unmuted':
					camAllowed = true;
					finishCamAndMicSetup();
				break;
			}
		}
		
		private function handleMicrophoneStatus(e:StatusEvent):void {
			trace("handleMicrophoneStatus - " + e.code);
			switch(e.code) {
				case 'Microphone.Muted':
					// Show a message
				break;
				case 'Microphone.Unmuted':
					micAllowed = true;
					finishCamAndMicSetup();
				break;
			}
		}
		
		private function handleBufferCheck(e:TimerEvent):void {
			if(activeStream != null) {
trace("Buffer: " + activeStream.bufferLength + " audio buffer: " + audioOnlyStream.bufferLength);
			
				if (recordingHalted == true) {
					if ( (activeStream.bufferLength == 0) && (audioOnlyStream.bufferLength == 0) ) {
						activeStream.close();
						audioOnlyStream.close();
						
						bufferCheckTimer.stop();
						bufferCheckTimer.removeEventListener(TimerEvent.TIMER, handleBufferCheck);
						bufferCheckTimer = null;
						
						// OK - playback time
						doRecordingPlayback();
					}
				}
			}
			
			if (bufferCheckTimer != null) {
				bufferCheckTimer.reset();
				bufferCheckTimer.start();
			}
		}
		
		private function handleAllClick(e:MouseEvent):void {
trace("Handle All Click - wait for the buffer to empty");
			if (activeStream != null) {

				activeStream.attachCamera(null);
				activeStream.attachAudio(null);
				
				audioOnlyStream.attachCamera(null);
				audioOnlyStream.attachAudio(null);
				
				recordingHalted = true;
				
				stage.removeEventListener(MouseEvent.CLICK, handleAllClick);
				// OK wait for the buffer to empty before closing
			}
		}
		
		// We're going to reuse our video and streams!
		private function doRecordingPlayback():void {
			onScreenVideo.attachCamera(null);
			onScreenVideo.attachNetStream(activeStream);
			audioPlaybackVideo.attachNetStream(audioOnlyStream);
			
			// Attempt to play back the audio and video file simultaniously
			// First, adjust the buffers as this will be BAD for playback
			activeStream.bufferTime = 1;
			audioOnlyStream.bufferTime = 1;
			
			activeStream.play(currentlyRecordedFileName + "_Video");
			audioOnlyStream.play(currentlyRecordedFileName + "_Audio");
		}
	}
	
}